#include <stdio.h>
#include <stdlib.h>

#define MAX_NUM 128
#define BUF_QUANTUM 104857600 /* 100M */

typedef unsigned long long uint64;

void touch(char *buf, int size) {
    int x;

    if (!buf)
         return;

    /* Quick-touch the "entire" buffer. */
    for (x = 0; x < size; x += 32)
        buf[x] = 0x5A;
}

/* Get used mem info from /proc/meminfo. */
uint64 getUsedMeminfo(const char *str) {
    FILE *fp = fopen("/proc/meminfo", "r");
    char line[128];
    uint64 total, used, free, meg;
    double gig;

    if (!fp)
        return;

    fgets(line, 128, fp);
    fscanf(fp, "Mem: %llu %llu %llu %s", &total, &used, &free, line);
    gig = used / 1024.0 / 1024.0 / 1024.0;
    meg = used / 1024uLL / 1024uLL;
    printf("%s  %llu (%lluM) (%.2fG)\n", str, used, meg, gig);

    fclose(fp);

    return (meg);
}

main(int argc, char *argv[]) {
    char **bufs;
    int size, num, x;
    uint64 before, after;

    if ((argc < 2) || !strcmp(argv[1], "-h")) {
        printf("usage:  %s size [num_bufs]\n\n"
               "        Specify size=0 (and omit num_bufs) for auto-reclamation.\n\n", argv[0]);
        return;
    }
    size = atoi(argv[1]);

    /* Performing auto-reclamation? */
    if (!size) {
        size = BUF_QUANTUM;
        num = 0;
        bufs = malloc(MAX_NUM * sizeof(char *));

        /* Get 'em */
        while (bufs[num++] = malloc(size));

        num--;
        printf("Allocated:      %d 100M buffers (%u)\n", num, num*size);
    } else {
        /* Number of buffers to allocate. */
        num = (argc > 2) ? atoi(argv[2]) : 1;

        /* Get 'em */
        bufs = malloc(num * sizeof(char *));
        for (x = 0; x < num; x++) {
            bufs[x] = malloc(size);
            printf("bufs[%d] malloc(%d) %sSUCCESSFUL (%08X)\n", x, size, bufs[x] ? "" : "UN", bufs[x]);
        }
    }

    before = getUsedMeminfo("Before Used:  ");

    /* Touch 'em */
    printf("Touching Buf:  ");
    for (x = 0; x < num; x++) {
        if (bufs[x]) {
            printf("%3d", x+1);
            fflush(stdout);
            touch(bufs[x], size);
            printf("");
        }
    }
    printf("\n");

    /* Free 'em */
    for (x = 0; x < num; x++) {
        if (bufs[x])
            free(bufs[x]);
    }
    free(bufs);

    after = getUsedMeminfo("After Used:   ");

    printf("Reclaimed:      %lluM\n", (before > after) ? (before - after) : 0uLL);
}

